#include <io.h>
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include "tags.h"

#define VERSION     "0.0.2"

typedef struct {
    tag_info    *tag;
    char        *file;
} cue_entry;

typedef struct {
    cue_entry   **track;
    size_t      tracks;
} cue_info;

// -------------------------------------

void clear ( cue_info *info )
{
    if ( info ) {
        for ( size_t i = 0; i < info->tracks; i++ ) {
            FreeTag ( info->track[i]->tag );
            if ( info->track[i]->tag  ) free ( info->track[i]->tag  );
            if ( info->track[i]->file ) free ( info->track[i]->file );
        }
        if ( info->track ) free ( info->track );
        info->track  = NULL;
        info->tracks = 0;
    }
}

// -------------------------------------

int read_tag ( const char *fn, tag_info *tag )
{
    FILE *fp = fopen ( fn, "rb" );
    if ( !fp ) {
        printf ( "Failed to open '%s'\n", fn );
        return -1;
    }

    int ret = ReadTag ( fp, tag );
    fclose ( fp );
    if ( ret != 0 ) {
        printf ( "Failed to read tags from '%s'\n", fn );
    }
    return ret;
}

int get_cue_info ( cue_info *info )
{
    struct _finddata_t  fd;
    long                hFile;

    memset ( info, 0, sizeof (cue_info) );

    hFile = _findfirst ( "*.mpc", &fd );
    if ( hFile == -1L )
        hFile = _findfirst ( "*.mp+", &fd );

    if ( hFile == -1L ) {
        printf ( "No Musepack files.\n" );
        return -1;
    }

    if ( !(fd.attrib & _A_SUBDIR || fd.attrib & _A_HIDDEN) ) {
        void *track_b = (void *)info->track;
        info->track = (cue_entry **)realloc ( info->track, (info->tracks+1) * sizeof (*info->track) );
        if ( !info->track ) {
            printf ( "Memory allocation failed.\n" );
            info->track = (cue_entry **)track_b;
            clear ( info );
            _findclose ( hFile );
            return -1;
        }

        info->track[info->tracks] = (cue_entry *)malloc ( sizeof (cue_entry) );
        if ( !info->track[info->tracks] ) {
            printf ( "Memory allocation failed.\n" );
            clear ( info );
            _findclose ( hFile );
            return -1;
        }

        info->track[info->tracks]->tag = (tag_info *)malloc ( sizeof (tag_info) );
        if ( !info->track[info->tracks]->tag ) {
            printf ( "Memory allocation failed.\n" );
            clear ( info );
            _findclose ( hFile );
            return -1;
        }

        if ( read_tag (fd.name, info->track[info->tracks]->tag) != 0 ) {
            FreeTag ( info->track[info->tracks]->tag );
            free ( info->track[info->tracks]->tag );
            clear ( info );
            _findclose ( hFile );
            return -1;
        }

        info->track[info->tracks]->file = strdup ( fd.name );
        if ( !info->track[info->tracks]->file ) {
            printf ( "Memory allocation failed.\n" );
            FreeTag ( info->track[info->tracks]->tag );
            free ( info->track[info->tracks]->tag );
            clear ( info );
            _findclose ( hFile );
            return -1;
        }

        info->tracks++;
    }

    while ( _findnext (hFile, &fd) == 0 ) {
        if ( !(fd.attrib & _A_SUBDIR || fd.attrib & _A_HIDDEN) ) {
            void *track_b = (void *)info->track;
            info->track = (cue_entry **)realloc ( info->track, (info->tracks+1) * sizeof (*info->track) );
            if ( !info->track ) {
                printf ( "Memory allocation failed.\n" );
                info->track = (cue_entry **)track_b;
                clear ( info );
                _findclose ( hFile );
                return -1;
            }

            info->track[info->tracks] = (cue_entry *)malloc ( sizeof (cue_entry) );
            if ( !info->track[info->tracks] ) {
                printf ( "Memory allocation failed.\n" );
                clear ( info );
                _findclose ( hFile );
                return -1;
            }

            info->track[info->tracks]->tag = (tag_info *)malloc ( sizeof (tag_info) );
            if ( !info->track[info->tracks]->tag ) {
                printf ( "Memory allocation failed.\n" );
                clear ( info );
                _findclose ( hFile );
                return -1;
            }

            if ( read_tag (fd.name, info->track[info->tracks]->tag) != 0 ) {
                FreeTag ( info->track[info->tracks]->tag );
                free ( info->track[info->tracks]->tag );
                clear ( info );
                _findclose ( hFile );
                return -1;
            }

            info->track[info->tracks]->file = strdup ( fd.name );
            if ( !info->track[info->tracks]->file ) {
                printf ( "Memory allocation failed.\n" );
                FreeTag ( info->track[info->tracks]->tag );
                free ( info->track[info->tracks]->tag );
                clear ( info );
                _findclose ( hFile );
                return -1;
            }

            info->tracks++;
        }
    }

    _findclose ( hFile );

    return 0;
}

// -------------------------------------

int cue_generate ( const char *name )
{
    const char *VA = "Various Artists";
    const char  *album_artist;
    const char  *album_title;
    cue_info    info;

    FILE *out = fopen ( name, "wt" );
    if ( !out ) {
        printf ( "Failed to create '%s'\n", name );
        return -1;
    }

    if ( get_cue_info (&info) != 0 ) {
        fclose ( out ); remove ( name );
        return -1;
    }

    if ( info.tracks == 0 ) {
        printf ( "No files?\n" );
        fclose ( out ); remove ( name );
        clear ( &info );
        return -1;
    }

    size_t i;
    album_title  = info.track[0]->tag->album;
    album_artist = info.track[0]->tag->artist;

    for ( i = 1; i < info.tracks; i++ ) {
        if ( strcmp (info.track[i]->tag->artist, info.track[i-1]->tag->artist) ) {
            album_artist = VA;
            break;
        }
    }

    bool faulty = false;

    for ( i = 0; i < info.tracks; i++ ) {
        if ( !info.track[i]->tag->track || atoi (info.track[i]->tag->track) <= 0 ) {
            faulty = true;
            break;
        }
    }

    if ( !faulty ) {
        for ( i = 0; i < info.tracks; i++ ) {
            int t1 = atoi ( info.track[i]->tag->track );
            for ( size_t j = 0; j < info.tracks; j++ ) {
                if ( i != j ) {
                    int t2 = atoi ( info.track[j]->tag->track );
                    if ( t1 == t2 ) {
                        faulty = true;
                        break;
                    }
                }
            }
            if ( faulty ) break;
        }
    }

    if ( faulty ) {
        printf ( "Invalid or missing track numbers.\n" );
        fclose ( out ); remove ( name );
        clear ( &info );
        return -1;
    }

    if ( album_artist && album_artist[0] )
        fprintf ( out, "PERFORMER \"%s\"\n", album_artist );
    if ( album_title && album_title[0] )
        fprintf ( out, "TITLE \"%s\"\n", album_title );

    for ( i = 0; i < info.tracks; i++ ) {
        cue_entry *p = NULL;
        int trackno = (int)(i+1);

        for ( size_t j = 0; j < info.tracks; j++ ) {
            if ( info.track[j]->tag->track ) {
                if ( atoi (info.track[j]->tag->track) == trackno ) {
                    p = info.track[j];
                    break;
                }
            }
        }

        if ( !p ) {
            printf ( "Invalid track numbers.\n" );
            fclose ( out ); remove ( name );
            clear ( &info );
            return -1;
        }

        fprintf ( out, "FILE \"%s\" WAVE\n", info.track[i]->file );
        fprintf ( out, "  TRACK %02u AUDIO\n", i+1 );
        if ( info.track[i]->tag->title && info.track[i]->tag->title[0] )
            fprintf ( out, "    TITLE \"%s\"\n", info.track[i]->tag->title );
        if ( info.track[i]->tag->artist && info.track[i]->tag->artist[0] )
            fprintf ( out, "    PERFORMER \"%s\"\n", info.track[i]->tag->artist );
        fprintf ( out, "    INDEX 01 00:00:00\n" );
    }

    fclose ( out );

    clear ( &info );

    printf ( "Done.\n" );

    return 0;
}

int main ( int argc, char **argv )
{
    if ( argc < 2 ) {
        printf ( "mpc2cue v" VERSION "\n" );
        printf ( "Usage: mpc2cue <cue name>\n" );
        return 0;
    }

    return cue_generate ( argv[1] );
}
